/* copyright (c)2012..2013 Oskar Schirmer <oskar@scara.com> */
#include <bfd.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

static int verbose = 0;

static void usage(char *arg0)
{
  fprintf(stderr,
    "Usage: %s <object file> <target file>\n"
    "transform elf object file to process target file\n"
    "\n"
    "  -f, --flags\t\tset core flags value (default: 0)\n"
    "  -s, --stack\t\tset stack size (default: 1024)\n"
    "  -h, --help\t\tshow this help\n"
    "  -v, --verbose\t\tprint various values to stderr\n"
    , arg0);
}

static unsigned align4(unsigned n)
{
  return (n + 3) & ~3;
}

static void writeword(int f, unsigned n)
{
  int i;
  for (i = 0; i < 4; i++) {
    unsigned char c;
    c = (unsigned char)n;
    if (write(f, &c, 1) != 1) {
      fprintf(stderr, "cannot write file\n");
      exit(1);
    }
    n >>= 8;
  }
}

static void writesection(int f, void *v, unsigned long n)
{
  int w, i = 0;
  while (i < n) {
    w = write(f, v + i, n - i);
    if (w <= 0) {
      fprintf(stderr, "cannot write file\n");
      exit(1);
    }
    i += w;
  }
}

static int nreloc_lu6(unsigned char *pc, unsigned value,
        unsigned op, unsigned mask, char *name)
{
  if (value & 3) {
    fprintf(stderr, "non aligned data reference\n");
    return 1;
  }
  value >>= 2;
  if (((pc[0] + (pc[1] << 8)) & mask) == op) {
    if (value >= 0x40) {
      fprintf(stderr, "value too large");
      return 1;
    }
    pc[0] += value;
    return 0;
  } else if ((pc[0] == 0x00) && (pc[1] == 0xf0)
          && (((pc[2] + (pc[3] << 8)) & mask) == op)) {
    if (value >= 0x10000) {
      fprintf(stderr, "value too large");
      return 1;
    }
    pc[0] = value >> 6;
    pc[1] += value >> 14;
    pc[2] += (value & 0x3f);
    return 0;
  }
  return 1;
}

static void *sectionalloc(bfd *b, asymbol **sym, asection *s,
        unsigned long v, unsigned long n, unsigned long *nn)
{
  void *m;
  int i, c;
  arelent **r, **p;
  if (v) {
    fprintf(stderr, "vma non zero, not implemented\n");
    exit(1);
  }
  m = malloc(*nn = align4(v + n));
  if (!bfd_get_section_contents(b, s, m + v, 0, n)) {
    fprintf(stderr, "cannot read section\n");
    exit(1);
  }
  if (s->flags & SEC_RELOC) {
    if (!(s->flags & SEC_CODE)) {
      fprintf(stderr, "data section reloc: not implemented\n");
      exit(1);
    }
    i = bfd_get_reloc_upper_bound(b, s);
    r = malloc(i);
    c = bfd_canonicalize_reloc(b, s, r, sym);
    if (verbose) {
      fprintf(stderr, "relocations: %d\n", c);
    }
    for (p = r; c && *p; p++, c--) {
      arelent *q = *p;
      struct bfd_symbol *y = *q->sym_ptr_ptr;
      if (verbose) {
        fprintf(stderr, "reloc %08x: %08x (%s)\n",
          (unsigned)q->address, (unsigned)y->value, y->name);
      }
      if (q->addend
       || q->howto->type || q->howto->rightshift || q->howto->size
       || q->howto->bitsize || q->howto->pc_relative || q->howto->bitpos
       || q->howto->partial_inplace || q->howto->pcrel_offset) {
        fprintf(stderr, "unexpected non zero reloc information\n");
        exit(1);
      }
      if (!(y->flags & BSF_LOCAL)) {
        fprintf(stderr, "symbol not local (%08x): %08x %s\n",
          (unsigned)q->address, (unsigned)y->value, y->name);
        exit(1);
      } else if (y->section->flags & SEC_DATA) {
        void *pc = m + v + q->address;
        unsigned val = y->value;
        if (y->section->flags & SEC_READONLY) {
          if (nreloc_lu6(pc, val, 0x7f40, 0xffc0, "LDAWCP")
           && nreloc_lu6(pc, val, 0x6c00, 0xfc00, "LDWCP")) {
            fprintf(stderr, "unknown relocation (%08x): %08x %s\n",
              (unsigned)q->address, (unsigned)y->value, y->name);
            exit(1);
          }
        } else {
          if (nreloc_lu6(pc, val, 0x6000, 0xfc00, "LDAWDP")
           && nreloc_lu6(pc, val, 0x5800, 0xfc00, "LDWDP")
           && nreloc_lu6(pc, val, 0x5000, 0xfc00, "STWDP")) {
            fprintf(stderr, "unknown relocation (%08x): %08x %s\n",
              (unsigned)q->address, (unsigned)y->value, y->name);
            exit(1);
          }
        }
      }
    }
  }
  return m;
}

static int nopc_lu6(unsigned char *pc, unsigned op, unsigned mask,
        long *value, int sign, char *name)
{
  unsigned v;
  if (((pc[1] & 0xfc) == 0xf0)
   && ((pc[2] + (pc[3] << 8)) & mask) == op) {
    v = (pc[2] & 0x3f) | ((((pc[1] & 3) << 8) | pc[0]) << 6);
    *value = v * sign;
    return 2;
  }
  if (((pc[0] + (pc[1] << 8)) & mask) == op) {
    v = pc[0] & 0x3f;
    *value = v * sign;
    return 1;
  }
  return 0;
}

static int nopc_lru6(unsigned char *pc, unsigned op, unsigned mask,
        long *value, int sign, char *name)
{
  unsigned v;
  if (((pc[1] & 0xfc) == 0xf0)
   && (((pc[2] + (pc[3] << 8)) & mask) == op)
   && ((((pc[2] + (pc[3] << 8)) >> 6) & 0x0f) < 12)) {
    v = (pc[2] & 0x3f) | ((((pc[1] & 3) << 8) | pc[0]) << 6);
    *value = v * sign;
    return 2;
  }
  if ((((pc[0] + (pc[1] << 8)) & mask) == op)
   && ((((pc[0] + (pc[1] << 8)) >> 6) & 0x0f) < 12)) {
    v = pc[0] & 0x3f;
    *value = v * sign;
    return 1;
  }
  return 0;
}

static int nopc_lu10(unsigned char *pc, unsigned op, unsigned mask,
        long *value, int sign, char *name)
{
  unsigned v;
  if (((pc[1] & 0xfc) == 0xf0)
   && ((pc[2] + (pc[3] << 8)) & mask) == op) {
    v = (((pc[3] & 3) << 8) | pc[2]) | ((((pc[1] & 3) << 8) | pc[0]) << 10);
    *value = v * sign;
    return 2;
  }
  if (((pc[0] + (pc[1] << 8)) & mask) == op) {
    v = (((pc[1] & 3) << 8) | pc[0]);
    *value = v * sign;
    return 1;
  }
  return 0;
}

static int nopc_any(unsigned char *pc, char *name)
{
  if ((pc[1] & 0xf0) == 0xf0) {
    return 2;
  }
  return 1;
}

enum codetype {fresh, code, prefixed, stack, call, entry};
struct stacksize {
  unsigned char typ;
  unsigned short func;
  unsigned value;
};

static void nprefixed(struct stacksize *s, int l)
{
  if (l == 2) {
    s[1].typ = prefixed;
  }
}

static int functionstacksize(unsigned char *v, unsigned long n,
        struct stacksize *s, unsigned long e, unsigned *func)
{
  int i, f;
  unsigned long x, a, y, z;
  long w;

  *func += 1;
  f = *func;
  z = 0;
  x = e;
  do {
    while ((x < n) && (s[x].typ <= code)) {
      s[x].func = f;
      if ((a = nopc_lu6(&v[2 * x], 0x7700, 0xffc0, &w, -1, "BRBU"))
       || (a = nopc_lu6(&v[2 * x], 0x7300, 0xffc0, &w, 1, "BRFU"))) {
        y = x + a + w;
        if (s[y].typ == fresh) {
          s[y].typ = code;
          s[y].value = z;
        }
        s[x].typ = stack;
        s[x].value = z;
        nprefixed(&s[x], a);
        x = n;
      } else
      if ((a = nopc_lru6(&v[2 * x], 0x7400, 0xf400, &w, -1, "BRBF/T"))
       || (a = nopc_lru6(&v[2 * x], 0x7000, 0xf400, &w, 1, "BRFF/T"))) {
        y = x + a + w;
        if (s[y].typ == fresh) {
          s[y].typ = code;
          s[y].value = z;
        }
        s[x].typ = stack;
        s[x].value = z;
        nprefixed(&s[x], a);
        x += a;
      } else
      if ((a = nopc_lu10(&v[2 * x], 0xd400, 0xfc00, &w, -1, "BLRB"))
       || (a = nopc_lu10(&v[2 * x], 0xd000, 0xfc00, &w, 1, "BLRF"))) {
        y = x + a + w;
        s[x].typ = call;
        s[x].value = y;
        nprefixed(&s[x], a);
        x += a;
      } else
      if ((a = nopc_lu6(&v[2 * x], 0x7740, 0xffc0, &w, 1, "ENTSP"))
       || (a = nopc_lu6(&v[2 * x], 0x7780, 0xffc0, &w, 1, "EXTSP"))) {
        z += w;
        s[x].typ = stack;
        s[x].value = z;
        nprefixed(&s[x], a);
        x += a;
      } else
      if (a = nopc_lu6(&v[2 * x], 0x77c0, 0xffc0, &w, -1, "RETSP")) {
        s[x].typ = stack;
        s[x].value = z;
        z += w;
        if (z != 0) {
          return -1;
        }
        nprefixed(&s[x], a);
        x = n;
      } else
      if (a = nopc_any(&v[2 * x], "other")) {
        s[x].typ = stack;
        s[x].value = z;
        nprefixed(&s[x], a);
        x += a;
      }
    }
    i = 1;
    for (x = 0; x < n; x++) {
      if (s[x].typ == code) {
        z = s[x].value;
        i = 0;
        break;
      }
    }
  } while (!i);
  y = 0;
  z = 0;
  for (x = 0; x < n; x++) {
    if (s[x].func == f) {
      if (s[x].typ == call) {
        w = s[x].value;
        if (s[w].typ == entry) {
          i = s[w].value;
        } else if (s[w].typ == fresh) {
          i = functionstacksize(v, n, s, w, func);
          if (i < 0) {
            return i;
          }
        } else {
          return -1;
        }
        if (y < i) {
          y = i;
        }
      } else {
        if (z < s[x].value) {
          z = s[x].value;
        }
      }
    }
  }
  s[e].typ = entry;
  s[e].value = y + z;
  if (verbose) {
    fprintf(stderr, "function %04x: stack %d+%d\n", e, y, z);
  }
  return y + z;
}

static int sectionstacksize(void *v, unsigned long n)
{
  struct stacksize *s;
  unsigned func;

  func = 0;
  s = malloc(n * sizeof(struct stacksize) / 2);
  memset(s, 0, n * sizeof(struct stacksize) / 2);
  return functionstacksize(v, n >> 1, s, 0, &func);
}

int main(int argc, char *argv[])
{
  static const struct option longopt[] = {
    {"flags", 1, 0, 'f'},
    {"stack", 1, 0, 's'},
    {"help", 0, 0, 'h'},
    {"verbose", 0, 0, 'v'},
    {0, 0, 0, 0}
  };
  int i, f;
  bfd *b;
  asymbol **sym;
  long nsym;
  asection *s;
  unsigned long v, n, npc, ncp, ndp;
  void *vpc, *vcp, *vdp;
  unsigned long *nn;
  void **vv;
  unsigned long flags, stack;

  flags = 0;
  stack = 0;
  while ((i = getopt_long(argc, argv, "f:s:hv", longopt, 0)) >= 0) {
    switch (i) {
    case 'f':
      flags = strtoul(optarg, 0, 0);
      break;
    case 's':
      stack = strtoul(optarg, 0, 0);
      break;
    case 'h':
      usage(argv[0]);
      exit(1);
    case 'v':
      verbose = 1;
      break;
    }
  }
  if ((optind + 2) > argc) {
    usage(argv[0]);
    exit(1);
  }
  if (!(b = bfd_openr(argv[optind], 0))) {
    fprintf(stderr, "cannot open %s\n", argv[optind]);
    exit(1);
  }
  if (!bfd_check_format(b, bfd_object)) {
    fprintf(stderr, "%s is not an object file\n", argv[optind]);
    exit(1);
  }
  if (b->arch_info->arch != bfd_arch_unknown) {
    fprintf(stderr, "input is not xmos architecture\n");
    exit(1);
  }
  if (!(bfd_get_file_flags(b) & HAS_RELOC)) {
    fprintf(stderr, "no relocations found\n");
    exit(1);
  }
  i = bfd_get_symtab_upper_bound(b);
  if (i > 0) {
    sym = malloc(i);
    nsym = bfd_canonicalize_symtab(b, sym);
  } else {
    nsym = 0;
  }
  if (verbose) {
    fprintf(stderr, "symbols: %ld (%d)\n", nsym, i);
  }
  i = 0;
  vpc = vcp = vdp = 0;
  npc = ncp = ndp = 0;
  for (s = b->sections; s; s = s->next) {
    v = bfd_section_vma(b, s);
    n = bfd_section_size(b, s);
    if (verbose) {
      fprintf(stderr,
        "id:%2d index:%2d flags:%08x vma:%08lx size:%08lx name:%s\n",
        s->id, s->index, s->flags, v, n, s->name);
    }
    if (s->flags & SEC_ALLOC) {
      if (!(s->flags & SEC_LOAD)) {
        fprintf(stderr, "bss section not allowed\n");
        i = 1;
      } else if (!(s->flags & SEC_HAS_CONTENTS)) {
        fprintf(stderr, "content free section not allowed\n");
        i = 1;
      } else if ((s->flags & SEC_CODE) && (s->flags & SEC_READONLY)
          && !strncmp(s->name, ".text", 5)) {
        if (npc) {
          fprintf(stderr, "double pc section\n");
          i = 1;
        }
        vpc = sectionalloc(b, sym, s, v, n, &npc);
      } else if ((s->flags & SEC_DATA) && (s->flags & SEC_READONLY)
          && !strncmp(s->name, ".cp", 3)) {
        if (ncp) {
          fprintf(stderr, "double cp section\n");
          i = 1;
        }
        vcp = sectionalloc(b, sym, s, v, n, &ncp);
      } else if ((s->flags & SEC_DATA) && !(s->flags & SEC_READONLY)
          && !strncmp(s->name, ".dp", 3)) {
        if (ndp) {
          fprintf(stderr, "double dp section\n");
          i = 1;
        }
        vdp = sectionalloc(b, sym, s, v, n, &ndp);
      } else {
        fprintf(stderr, "unexpected loaded content section\n");
        i = 1;
      }
    }
  }
  if (i) {
    exit(1);
  }
  if (!npc) {
    fprintf(stderr, "no pc section\n");
    exit(1);
  }
  i = sectionstacksize(vpc, npc) * 4;
  if (i < 0) {
    fprintf(stderr, "warning: recursivness\n");
  } else if (stack) {
    if (stack < i) {
      fprintf(stderr, "warning: stack too small\n");
    }
  } else {
    stack = i;
  }
  if (!(f = open(argv[optind + 1], O_WRONLY | O_CREAT | O_TRUNC, 0744))) {
    fprintf(stderr, "cannot open %s\n", argv[optind + 1]);
    exit(1);
  }
  writeword(f, 0xbece);
  writeword(f, flags);
  writeword(f, 16 + npc + ncp + ndp);
  writeword(f, stack);
  writeword(f, 12);
  writeword(f, 8 + npc);
  writeword(f, 4 + npc + ncp);
  writesection(f, vpc, npc);
  writesection(f, vcp, ncp);
  writesection(f, vdp, ndp);
  close(f);
}
